import typing as t

if t.TYPE_CHECKING:
    from zndraw.base import Extension


class RegisterModifier(t.TypedDict):
    cls: t.Type["Extension"]
    run_kwargs: dict
    public: bool


class TimeoutConfig(t.TypedDict):
    """Timeout configuration for the ZnDraw client."""

    connection: int
    modifier: float
    between_calls: float

    emit_retries: int
    call_retries: int
    connect_retries: int


class JupyterConfig(t.TypedDict):
    width: str | int
    height: str | int


class CameraData(t.TypedDict):
    position: list[float]
    target: list[float]


class ASEDict(t.TypedDict):
    numbers: list[int]
    positions: list[list[float]]
    connectivity: list[tuple[int, int, int]]
    arrays: dict[str, list[float | int | list[float | int]]]
    info: dict[str, float | int]
    # calc: dict[str, float|int|np.ndarray] # should this be split into arrays and info?
    pbc: list[bool]
    cell: list[list[float]]
    vectors: list[list[list[float]]]
    constraints: list[dict]


class ASEJson(t.TypedDict):
    _type: t.Literal["ase.Atoms"]
    value: ASEDict


# Type hint is string, but correctly it is 'json.dumps(ASEJson)'
ATOMS_LIKE = t.Union[ASEDict, str]



import dataclasses
import logging
import typing as t
from abc import abstractmethod

from pydantic import BaseModel

if t.TYPE_CHECKING:
    from zndraw import ZnDraw

log = logging.getLogger(__name__)


class Extension(BaseModel):
    @abstractmethod
    def run(self, vis: "ZnDraw", **kwargs) -> None:
        raise NotImplementedError("run method must be implemented in subclass")


@dataclasses.dataclass  # TODO: move to a separate file, so it can be imported in other files
class FileIO:
    name: str | None = None
    start: int = 0
    stop: int | None = None
    step: int = 1
    remote: str | None = None
    rev: str | None = None
    convert_nan: bool = False

    def to_dict(self):
        return dataclasses.asdict(self)



import base64
import dataclasses

import matplotlib.pyplot as plt
import znjson


@dataclasses.dataclass(frozen=True)
class Figure:
    """Visualize a file or a matplotlib figure."""

    path: str | None = None
    figure: plt.Figure | None = None

    def __post_init__(self):
        if self.path is not None and self.figure is not None:
            raise ValueError("Figure can't have both path and figure")

    def to_base64(self) -> str:
        if self.path is not None:
            with open(self.path, "rb") as image_file:
                return base64.b64encode(image_file.read()).decode("utf-8")
        else:
            import io

            buf = io.BytesIO()
            self.figure.savefig(buf, format="png")
            buf.seek(0)
            return base64.b64encode(buf.read()).decode("utf-8")


class FigureConverter(znjson.ConverterBase):
    level = 100
    instance = Figure
    representation = "zndraw.Figure"

    def encode(self, obj: Figure) -> dict:
        return {"path": obj.path, "base64": obj.to_base64()}

    def decode(self, value: dict) -> Figure:
        return Figure(value["path"])



class ZnDrawException(Exception):
    """Base ZnDraw exception."""


class RoomNotFound(Exception):
    """Raised when a room is not found on the server."""


class RoomLockedError(ZnDrawException):
    """Raised when tried to modify a locked room."""



"""zndraw-upload API"""

import typing as t
import uuid
import webbrowser

import typer

from zndraw import ZnDraw

from .tasks import FileIO, get_generator_from_filename
from .utils import load_plots_to_dict

cli = typer.Typer()


def upload(
    url: str,
    token: t.Optional[str],
    fileio: FileIO,
    append: bool,
    plots: list[str],
    browser: bool,
    batch_size: int = 16,
):
    """Upload a file to ZnDraw."""
    if token is None:
        token = str(uuid.uuid4())
    vis = ZnDraw(url=url, token=token, convert_nan=fileio.convert_nan)
    typer.echo(f"Uploading to: {url}/token/{vis.token}")

    if not append:
        del vis[:]
    typer.echo(f"Reading {fileio.name} ...")

    generator = get_generator_from_filename(fileio)

    if browser:
        webbrowser.open(f"{url}/token/{vis.token}")

    frames = []
    for frame in generator:
        frames.append(frame)
        if len(frames) == batch_size:
            vis.extend(frames)
            frames = []
    vis.extend(frames)

    vis.figures.update(load_plots_to_dict(plots, fileio.remote, fileio.rev))



import logging
import typing as t
import urllib.request
from io import StringIO

import ase.io
import znsocket
from celery import shared_task
from flask import current_app

from zndraw.analyse import analyses
from zndraw.base import FileIO
from zndraw.bonds import ASEComputeBonds
from zndraw.config import SETTINGS
from zndraw.draw import geometries
from zndraw.modify import modifier
from zndraw.queue import run_queued_task
from zndraw.selection import selections
from zndraw.utils import load_plots_to_dict

log = logging.getLogger(__name__)


def _get_default_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    return [ase.Atoms()]


def _get_zntrack_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    node_name, attribute = file_io.name.split(".", 1)
    try:
        import os

        ## FIX for zntrack bug https://github.com/zincware/ZnTrack/issues/806
        import sys

        import zntrack

        sys.path.insert(0, os.getcwd())
        ##

        node = zntrack.from_rev(node_name, remote=file_io.remote, rev=file_io.rev)
        images = node
        for key in attribute.split("."):
            if isinstance(images, list):
                images = images[int(key)]
                if not isinstance(images, list):
                    images = [images]
            else:
                images = getattr(images, key)

    except ImportError as err:
        raise ImportError(
            "You need to install ZnTrack to use the remote feature."
        ) from err

    return images[file_io.start : file_io.stop : file_io.step]


def _get_znh5md_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    try:
        import znh5md

        io = znh5md.IO(file_io.name)
    except ImportError as err:
        raise ImportError(
            "You need to install ZnH5MD>=0.3 to use the remote feature."
        ) from err
    log.critical(file_io)
    return io[file_io.start : file_io.stop : file_io.step]


def _get_http_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    format = file_io.name.split(".")[-1]
    format = format if format != "xyz" else "extxyz"
    content = urllib.request.urlopen(file_io.name).read().decode("utf-8")
    stream = StringIO(content)

    generator = ase.io.iread(
        stream, format=format, index=slice(file_io.start, file_io.stop, file_io.step)
    )

    return generator


def _get_ase_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    generator = ase.io.iread(
        file_io.name, index=slice(file_io.start, file_io.stop, file_io.step)
    )

    return generator


def get_generator_from_filename(
    file_io: FileIO, bond_calculator: ASEComputeBonds | None = None
) -> t.Iterable[ase.Atoms]:
    if file_io.name is None:
        gen = _get_default_generator(file_io)
    elif file_io.remote is not None:
        gen = _get_zntrack_generator(file_io)
    elif file_io.name.endswith((".h5", ".hdf5", ".h5md")):
        gen = _get_znh5md_generator(file_io)
    elif file_io.name.startswith(("http", "https")):
        gen = _get_http_generator(file_io)
    else:
        gen = _get_ase_generator(file_io)

    if bond_calculator is not None:
        for atoms in gen:
            if not hasattr(atoms, "connectivity"):
                atoms.connectivity = bond_calculator.get_bonds(atoms)
            yield atoms
    else:
        yield from gen


@shared_task
def read_file(fileio: dict) -> None:
    file_io = FileIO(**fileio)

    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token="default",
        convert_nan=file_io.convert_nan,
    )
    bonds_calculator = ASEComputeBonds()
    if current_app.config.get("COMPUTE_BONDS", False):
        generator = get_generator_from_filename(file_io, bonds_calculator)
    else:
        generator = get_generator_from_filename(file_io)

    # TODO: vis.extend(generator) # vis does not yet support consuming a generator
    atoms_buffer = []
    for atoms in generator:
        atoms_buffer.append(atoms)
        if len(atoms_buffer) > 10:
            vis.extend(atoms_buffer)
            atoms_buffer = []
    vis.extend(atoms_buffer)

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_upload_file(room, data: dict):
    from io import StringIO

    import ase.io

    from zndraw.zndraw import ZnDrawLocal

    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )

    vis.log(f"Uploading {data['filename']}")

    format = data["filename"].split(".")[-1]
    format = format if format != "xyz" else "extxyz"

    if format == "h5":
        raise ValueError("H5MD format not supported for uploading yet")

    stream = StringIO(bytes(data["content"]).decode("utf-8"))

    atoms_list = list(ase.io.iread(stream, format=format))
    if len(atoms_list) == 1 and len(vis.points) != 0:
        scene = vis.atoms
        atoms = atoms_list[0]
        if hasattr(scene, "connectivity"):
            del scene.connectivity
        for point in vis.points:
            atoms.positions -= atoms.get_center_of_mass() - point
            scene += atoms
        vis.append(scene)
    else:
        vis.extend(atoms_list)

    vis.step = len(vis) - 1

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def read_plots(paths: list[str], remote, rev) -> None:
    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token="default",
    )

    vis.figures = load_plots_to_dict(paths, remote, rev)

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def load_zntrack_frames(room: str, remote: str, rev: str, attribute: str, name: str):
    if len(rev) == 0:
        rev = None
    import zntrack

    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )
    try:
        vis.extend(getattr(zntrack.from_rev(name, remote=remote, rev=rev), attribute))
    except Exception as e:
        vis.log(str(e))

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def load_zntrack_figures(room: str, remote: str, rev: str, attribute: str, name: str):
    if len(rev) == 0:
        rev = None
    import zntrack

    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )
    try:
        vis.figures.update(
            getattr(zntrack.from_rev(name, remote=remote, rev=rev), attribute)
        )
    except Exception as e:
        vis.log(str(e))

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def inspect_zntrack_node(name, rev, remote):
    import dataclasses

    import zntrack
    from zntrack.config import ZNTRACK_OPTION, ZnTrackOptionEnum

    node = zntrack.from_rev(name=name, remote=remote, rev=rev)

    # find all @property decorated methods
    def find_properties(cls):
        exclude = set(
            [
                "_init_descriptors_",
                "_init_subclass_basecls_",
                "_use_repr_",
                "nwd",
                "state",
                "uuid",
            ]
        )
        names = []
        for name in dir(cls):
            if name not in exclude:
                attr = getattr(cls, name)
                if isinstance(attr, property):
                    # get the type hint of the property
                    hint = attr.fget.__annotations__.get("return", None)
                    names.append((name, str(hint)))
        return names

    def find_deps_outs(cls):
        names = []
        for field in dataclasses.fields(cls):
            if field.metadata.get(ZNTRACK_OPTION) in [
                ZnTrackOptionEnum.OUTS,
                ZnTrackOptionEnum.DEPS,
            ]:
                # names.append(field.name)
                # append a tuple with the name and the type
                names.append((field.name, str(field.type)))
        return names

    return find_deps_outs(node) + find_properties(node.__class__)


@shared_task
def setup_public_modifier(room):
    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )

    room_modifier_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:modifier",
    )

    # We link the room queue to the default queue
    # so that the public modifier can be used in the room
    # public modifiers iterate the default_modifier_queue
    # and check for applicable tasks to pick up.
    # TODO: remove rooms from this list if they are closed
    # TODO: when to close a room?
    default_modifier_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key="queue:default:modifier",
    )

    default_modifier_queue[room] = room_modifier_queue

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_room_worker(room):
    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )

    geometry_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:geometry",
    )
    for key in geometry_queue:
        if key in geometries:
            try:
                task = geometry_queue.pop(key)
                vis.geometries.append(geometries[key](**task))
            except IndexError:
                pass

    selection_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:selection",
    )

    for key in selection_queue:
        if key in selections:
            try:
                task = selection_queue.pop(key)
                run_queued_task(vis, selections[key], task, selection_queue)
            except IndexError:
                pass

    analysis_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:analysis",
    )

    for key in analysis_queue:
        if key in analyses:
            try:
                task = analysis_queue.pop(key)
                try:
                    run_queued_task(vis, analyses[key], task, analysis_queue)
                except Exception as err:
                    vis.log(f"Error running analysis `{key}`: {err}")
            except IndexError:
                pass

    scene_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:settings",
    )

    for key, val in scene_queue.items():
        vis.config[key].update(val)
        # TODO: also update the schema to update all other rooms
        vis.config["trigger_update"] = True

    modifier_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:modifier",
    )

    for key in modifier_queue:
        if key in modifier:
            try:
                task = modifier_queue.pop(key)
                try:
                    run_queued_task(vis, modifier[key], task, modifier_queue)
                except Exception as err:
                    vis.log(f"Error running modifier `{key}`: {err}")
            except IndexError:
                pass

    # wait and then disconnect
    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_schema(room) -> None:
    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )

    room_modifier_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"queue:{room}:modifier",
    )

    # We link the room queue to the default queue
    # so that the public modifier can be used in the room
    # public modifiers iterate the default_modifier_queue
    # and check for applicable tasks to pick up.
    # TODO: remove rooms from this list if they are closed
    # TODO: when to close a room?
    default_modifier_queue = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key="queue:default:modifier",
    )

    default_modifier_queue[room] = room_modifier_queue

    dct = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key="schema:default:modifier",
    )
    for key, val in modifier.items():
        dct[key] = val.model_json_schema()

    dct = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"schema:{room}:selection",
    )
    for key, val in selections.items():
        dct[key] = val.model_json_schema()

    dct = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"schema:{room}:geometry",
    )
    for key, val in geometries.items():
        dct[key] = val.model_json_schema()

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_scene_dependent_schema(room) -> None:
    from zndraw import ZnDraw

    vis = ZnDraw(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )

    dct = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"schema:{room}:analysis",
    )
    for key, val in analyses.items():
        dct[key] = val.model_json_schema_from_atoms(vis.atoms)

    dct = znsocket.Dict(
        r=current_app.extensions["redis"],
        socket=vis._refresh_client,
        key=f"schema:{room}:settings",
    )

    for key, val in SETTINGS.items():
        config_values = dict(vis.config[key])
        schema = val.model_json_schema_from_atoms(vis.atoms)
        for prop, val in schema["properties"].items():
            try:
                schema["properties"][prop]["default"] = config_values[prop]
            except KeyError:
                vis.log(f"KeyError: {prop}")
        dct[key] = schema

    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_room_copy(room) -> None:
    # check if the room exists or if default room is used instead
    r = current_app.extensions["redis"]

    if not r.exists(f"room:{room}:frames") and r.exists("room:default:frames"):
        default_lst = znsocket.List(
            r,
            "room:default:frames",
        )
        if not (len(default_lst) == 1 and len(default_lst[0]) == 0):
            # prevent copying empty default room
            default_lst.copy(key=f"room:{room}:frames")



import typing as t

from pydantic import BaseModel, ConfigDict, Field


def _update_material_schema(schema: dict) -> dict:
    schema["properties"]["wireframe"]["format"] = "checkbox"
    schema["properties"]["color"]["format"] = "color"
    schema["properties"]["opacity"]["format"] = "range"
    schema["properties"]["opacity"]["step"] = 0.01
    schema["properties"]["outlines"]["format"] = "checkbox"
    return schema


def _update_object3d_schema(schema: dict) -> dict:
    """Remote position, rotation, and scale from the schema."""
    schema["properties"].pop("position", None)
    schema["properties"].pop("rotation", None)
    schema["properties"].pop("scale", None)
    return schema


class Material(BaseModel):
    color: str = "#62929E"
    opacity: float = Field(default=0.2, ge=0.0, le=1.0)
    wireframe: bool = False
    outlines: bool = False

    model_config = ConfigDict(json_schema_extra=_update_material_schema)


class Object3D(BaseModel):
    """Base class for all 3D objects."""

    material: Material = Material()

    position: t.Tuple[float, float, float] | list[float] = (0, 0, 0)
    rotation: t.Tuple[float, float, float] | list[float] = (0, 0, 0)
    scale: t.Tuple[float, float, float] | list[float] = (1, 1, 1)

    model_config = ConfigDict(json_schema_extra=_update_object3d_schema)

    def run(self, vis, **kwargs) -> None:
        # get the selected particles and compute the COM
        if len(vis.selection) > 0:
            selected = vis.atoms[vis.selection]
            self.position = selected.get_center_of_mass().tolist()
        # self.position = vis.points[0].tolist()
        print(f"Running {self.__class__.__name__} at {self.position}")

        vis.geometries.append(self)  # TODO: dump / load without pickle


class Plane(Object3D):
    width: float = 10.0
    height: float = 10.0


class Box(Object3D):
    width: float = 10.0
    height: float = 10.0
    depth: float = 10.0


class Circle(Object3D):
    radius: float = 5.0


class Cone(Object3D):
    radius: float = 5.0
    height: float = 10.0


class Cylinder(Object3D):
    radius_top: float = 5.0
    radius_bottom: float = 5.0
    height: float = 10.0


class Dodecahedron(Object3D):
    radius: float = 5.0


class Icosahedron(Object3D):
    radius: float = 5.0


class Octahedron(Object3D):
    radius: float = 5.0


class Ring(Object3D):
    inner_radius: float = 1
    outer_radius: float = 4.0


class Sphere(Object3D):
    radius: float = 4.0


class Tetrahedron(Object3D):
    radius: float = 5.0


class Torus(Object3D):
    radius: float = 3.0
    tube: float = 1.0


class TorusKnot(Object3D):
    radius: float = 3.0
    tube: float = 1.0


class Rhomboid(Object3D):
    vectorA: t.Tuple[float, float, float] | list[float] = (10, 0, 0)
    vectorB: t.Tuple[float, float, float] | list[float] = (0, 10, 0)
    vectorC: t.Tuple[float, float, float] | list[float] = (0, 0, 10)


class Custom2DShape(Object3D):
    points: list[tuple[float, float]]


class Ellipsoid(Object3D):
    a: float = 10.0
    b: float = 5.0
    c: float = 5.0


geometries: dict[str, t.Type[Object3D]] = {
    Plane.__name__: Plane,
    Box.__name__: Box,
    Circle.__name__: Circle,
    Cone.__name__: Cone,
    Cylinder.__name__: Cylinder,
    Dodecahedron.__name__: Dodecahedron,
    Icosahedron.__name__: Icosahedron,
    Octahedron.__name__: Octahedron,
    Ring.__name__: Ring,
    Sphere.__name__: Sphere,
    Tetrahedron.__name__: Tetrahedron,
    Torus.__name__: Torus,
    TorusKnot.__name__: TorusKnot,
    Rhomboid.__name__: Rhomboid,
    Ellipsoid.__name__: Ellipsoid,
}



import typing as t

import ase

if t.TYPE_CHECKING:
    from zndraw.zndraw import ZnDraw

from . import UpdateScene


class NewScene(UpdateScene):
    """Clear the scene, deleting all atoms and points."""

    def run(self, vis: "ZnDraw", **kwargs) -> None:
        del vis[vis.step + 1 :]
        vis.points = []
        vis.append(ase.Atoms())
        vis.selection = []
        step = len(vis) - 1
        vis.step = step
        vis.bookmarks = vis.bookmarks | {step: "New Scene"}
        vis.camera = {"position": [0, 0, 20], "target": [0, 0, 0]}


class ClearTools(UpdateScene):
    """Clear the tools, removing all guiding points and undoing any selection."""

    def run(self, vis: "ZnDraw", **kwargs) -> None:
        vis.points = []
        vis.selection = []



import ase
import numpy as np
import znjson
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.data.colors import jmol_colors

from zndraw.draw import Object3D
from zndraw.type_defs import ASEDict
from zndraw.utils import get_scaled_radii, rgb2hex


class ASEConverter(znjson.ConverterBase):
    """Encode/Decode datetime objects

    Attributes
    ----------
    level: int
        Priority of this converter over others.
        A higher level will be used first, if there
        are multiple converters available
    representation: str
        An unique identifier for this converter.
    instance:
        Used to select the correct converter.
        This should fulfill isinstance(other, self.instance)
        or __eq__ should be overwritten.
    """

    level = 100
    representation = "ase.Atoms"
    instance = ase.Atoms

    def encode(self, obj: ase.Atoms) -> ASEDict:
        """Convert the datetime object to str / isoformat"""

        numbers = obj.numbers.tolist()
        positions = obj.positions.tolist()
        pbc = obj.pbc.tolist()
        cell = obj.cell.tolist()

        info = {
            k: v
            for k, v in obj.info.items()
            if isinstance(v, (float, int, str, bool, list))
        }
        info |= {k: v.tolist() for k, v in obj.info.items() if isinstance(v, np.ndarray)}
        vectors = info.pop("vectors", [])
        if isinstance(vectors, np.ndarray):
            vectors = vectors.tolist()
        for idx, vector in enumerate(vectors):
            if isinstance(vector, np.ndarray):
                vectors[idx] = vector.tolist()

        if len(vectors) != 0:
            vectors = np.array(vectors)
            if vectors.ndim != 3:
                raise ValueError(
                    f"Vectors must be of shape (n, 2, 3), found '{vectors.shape}'"
                )
            if vectors.shape[1] != 2:
                raise ValueError(
                    f"Vectors must be of shape (n, 2, 3), found '{vectors.shape}'"
                )
            if vectors.shape[2] != 3:
                raise ValueError(
                    f"Vectors must be of shape (n, 2, 3), found '{vectors.shape}'"
                )

            vectors = vectors.tolist()

        if obj.calc is not None:
            calc = {
                k: v
                for k, v in obj.calc.results.items()
                if isinstance(v, (float, int, str, bool, list))
            }
            calc |= {
                k: v.tolist()
                for k, v in obj.calc.results.items()
                if isinstance(v, np.ndarray)
            }
        else:
            calc = {}

        # All additional information should be stored in calc.results
        # and not in calc.arrays, thus we will not convert it here!
        arrays = {}
        if ("colors" not in obj.arrays) or ("" in obj.arrays["colors"]):
            arrays["colors"] = [rgb2hex(jmol_colors[number]) for number in numbers]
        else:
            arrays["colors"] = (
                obj.arrays["colors"].tolist()
                if isinstance(obj.arrays["colors"], np.ndarray)
                else obj.arrays["colors"]
            )

        if ("radii" not in obj.arrays) or (0 in obj.arrays["radii"]):
            # arrays["radii"] = [covalent_radii[number] for number in numbers]
            arrays["radii"] = [get_scaled_radii()[number] for number in numbers]
        else:
            arrays["radii"] = (
                obj.arrays["radii"].tolist()
                if isinstance(obj.arrays["radii"], np.ndarray)
                else obj.arrays["radii"]
            )

        for key in obj.arrays:
            if key in ["colors", "radii"]:
                continue
            if isinstance(obj.arrays[key], np.ndarray):
                arrays[key] = obj.arrays[key].tolist()
            else:
                arrays[key] = obj.arrays[key]

        if hasattr(obj, "connectivity") and obj.connectivity is not None:
            connectivity = (
                obj.connectivity.tolist()
                if isinstance(obj.connectivity, np.ndarray)
                else obj.connectivity
            )
        else:
            connectivity = []

        constraints = []
        if len(obj.constraints) > 0:
            for constraint in obj.constraints:
                if isinstance(constraint, FixAtoms):
                    constraints.append(
                        {"type": "FixAtoms", "indices": constraint.index.tolist()}
                    )
                else:
                    # Can not serialize other constraints
                    pass

        # We don't want to send positions twice
        arrays.pop("positions", None)
        arrays.pop("numbers", None)

        return ASEDict(
            numbers=numbers,
            positions=positions,
            connectivity=connectivity,
            arrays=arrays,
            info=info,
            calc=calc,
            pbc=pbc,
            cell=cell,
            vectors=vectors,
            constraints=constraints,
        )

    def decode(self, value: ASEDict) -> ase.Atoms:
        """Create datetime object from str / isoformat"""
        atoms = ase.Atoms(
            numbers=value["numbers"],
            positions=value["positions"],
            info=value["info"],
            pbc=value["pbc"],
            cell=value["cell"],
        )
        if connectivity := value.get("connectivity"):
            # or do we want this to be nx.Graph?
            atoms.connectivity = np.array(connectivity)
        for key, val in value["arrays"].items():
            atoms.arrays[key] = np.array(val)
        if calc := value.get("calc"):
            atoms.calc = SinglePointCalculator(atoms)
            atoms.calc.results.update(calc)
        if vectors := value.get("vectors"):
            atoms.info["vectors"] = vectors
        if constraints := value.get("constraints"):
            for constraint in constraints:
                if constraint["type"] == "FixAtoms":
                    atoms.set_constraint(FixAtoms(constraint["indices"]))
        return atoms


class Object3DConverter(znjson.ConverterBase):
    instance: type = Object3D
    representation: str = "zndraw.Object3D"
    level: int = 100

    def encode(self, obj: Object3D) -> dict:
        return {"class": obj.__class__.__name__, "data": obj.model_dump()}

    def decode(self, value: str) -> Object3D:
        import zndraw

        cls = getattr(zndraw, value["class"])

        return cls(**value["data"])



import enum
import typing as t

import ase
import numpy as np
from pydantic import BaseModel, Field

if t.TYPE_CHECKING:
    pass


HSLColor = t.Tuple[float, float, float]


class SettingsBase(BaseModel):
    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        return cls.model_json_schema()


class Material(str, enum.Enum):
    MeshBasicMaterial = "MeshBasicMaterial"
    # MeshLambertMaterial = "MeshLambertMaterial"
    # MeshMatcapMaterial = "MeshMatcapMaterial"
    # MeshPhongMaterial = "MeshPhongMaterial"
    MeshPhysicalMaterial = "MeshPhysicalMaterial"
    MeshStandardMaterial = "MeshStandardMaterial"
    MeshToonMaterial = "MeshToonMaterial"


class Controls(str, enum.Enum):
    OrbitControls = "OrbitControls"
    TrackballControls = "TrackballControls"


class CameraEnum(str, enum.Enum):
    PerspectiveCamera = "PerspectiveCamera"
    OrthographicCamera = "OrthographicCamera"


class Particle(SettingsBase):
    particle_size: float = Field(1.0, ge=0.1, le=5, description="Particle Size")
    bond_size: float = Field(1.0, ge=0.1, le=5, description="Bonds Size")
    material: Material = Field(Material.MeshStandardMaterial, description="Material")
    selection_color: str = Field("#ffa500", description="Selection color")
    hover_opacity: float = Field(0.8, ge=0.0, le=1.0, description="Hover opacity")
    selection_opacity: float = Field(0.5, ge=0.0, le=1.0, description="Selection opacity")

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()
        schema["properties"]["particle_size"]["format"] = "range"
        schema["properties"]["particle_size"]["step"] = 0.1
        schema["properties"]["bond_size"]["format"] = "range"
        schema["properties"]["bond_size"]["step"] = 0.1
        schema["properties"]["selection_color"]["format"] = "color"
        schema["properties"]["hover_opacity"]["format"] = "range"
        schema["properties"]["hover_opacity"]["step"] = 0.05
        schema["properties"]["selection_opacity"]["format"] = "range"
        schema["properties"]["selection_opacity"]["step"] = 0.05
        return schema


class VectorDisplay(SettingsBase):
    vectorfield: bool = Field(True, description="Show vectorfield.")
    vectors: list = Field("", description="Visualize vectorial property")
    vector_scale: float = Field(1.0, ge=0.1, le=5, description="Rescale Vectors")

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()
        array_props = []
        if atoms.calc is not None:
            for key in atoms.calc.results.keys():
                if (
                    np.array(atoms.calc.results[key]).ndim == 2
                    and np.array(atoms.calc.results[key]).shape[1] == 3
                ):
                    array_props.append(key)
        for key in atoms.arrays.keys():
            if (
                np.array(atoms.arrays[key]).ndim == 2
                and np.array(atoms.arrays[key]).shape[1] == 3
            ):
                array_props.append(key)
        # remove "positions" from the list
        array_props = [x for x in array_props if x != "positions"]
        schema["properties"]["vectors"]["items"] = {"type": "string", "enum": array_props}
        schema["properties"]["vectors"]["uniqueItems"] = True
        schema["properties"]["vectorfield"]["format"] = "checkbox"
        schema["properties"]["vector_scale"]["format"] = "range"
        schema["properties"]["vector_scale"]["step"] = 0.05
        return schema


class Visualization(SettingsBase):
    simulation_box: bool = Field(
        True,
        description="Show the simulation box.",
    )
    floor: bool = Field(False, description="Show the floor.")
    frame_update: bool = Field(
        True,
        description="Jump to updated frames.",
    )
    animation_loop: bool = Field(
        False,
        description="Automatically restart animation when finished.",
    )

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()
        schema["properties"]["simulation_box"]["format"] = "checkbox"
        schema["properties"]["frame_update"]["format"] = "checkbox"
        schema["properties"]["animation_loop"]["format"] = "checkbox"
        schema["properties"]["floor"]["format"] = "checkbox"
        return schema


class Camera(SettingsBase):
    camera: CameraEnum = Field(CameraEnum.PerspectiveCamera)
    camera_near: float = Field(
        0.1, ge=0, le=100, description="Camera near rendering plane"
    )
    camera_far: float = Field(
        300, ge=1, le=1000, description="Camera far rendering plane"
    )
    crosshair: bool = Field(
        False,
        description="Show camera controls target.",
    )
    synchronize_camera: bool = Field(
        True,
        description="Synchronize camera with other room members.",
    )
    fps: int = Field(30, ge=1, le=120, description="Maximum frames per second")
    controls: Controls = Field(Controls.OrbitControls, description="Controls")

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()
        schema["properties"]["fps"]["format"] = "range"
        schema["properties"]["fps"]["step"] = 1
        schema["properties"]["camera_near"]["format"] = "range"
        schema["properties"]["camera_near"]["step"] = 0.1
        schema["properties"]["camera_far"]["format"] = "range"
        schema["properties"]["camera_far"]["step"] = 1

        schema["properties"]["crosshair"]["format"] = "checkbox"
        schema["properties"]["synchronize_camera"]["format"] = "checkbox"
        return schema


class Arrows(SettingsBase):
    """Experimental vector color settings."""

    colormap: list[HSLColor] = ((0.5, 0.9, 0.5), (1.0, 0.9, 0.5))
    normalize: bool = True
    colorrange: tuple[float, float] = (0, 1.0)
    scale_vector_thickness: bool = False
    opacity: float = Field(1.0, ge=0.0, le=1.0, description="Opacity")

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()
        schema["properties"]["colormap"] = {
            "type": "array",
            "description": "Defines the colormap as a list of HSL tuples (Hue, Saturation, Lightness).",
            "items": {
                "type": "array",
                "minItems": 3,
                "maxItems": 3,
                "headertemplate": "{{ i1 }}",
                "items": [
                    {
                        "type": "number",
                        "description": "Hue (0.0 - 1.0)",
                        "format": "range",
                        "step": 0.01,
                        "minimum": 0,
                        "maximum": 1,
                    },
                    {
                        "type": "number",
                        "description": "Saturation (0.0 - 1.0)",
                        "format": "range",
                        "step": 0.01,
                        "minimum": 0,
                        "maximum": 1,
                    },
                    {
                        "type": "number",
                        "description": "Lightness (0.0 - 1.0)",
                        "format": "range",
                        "step": 0.01,
                        "minimum": 0,
                        "maximum": 1,
                    },
                ],
            },
        }

        # Enhance "colorrange"
        schema["properties"]["colorrange"] = {
            "type": "array",
            "description": "Specifies the range of values for colors, defined as [min, max].",
            "minItems": 2,
            "maxItems": 2,
            "items": [
                {
                    "type": "number",
                    "description": "Minimum value of the range.",
                    "format": "range",
                    "step": 0.01,
                },
                {
                    "type": "number",
                    "description": "Maximum value of the range.",
                    "format": "range",
                    "step": 0.01,
                },
            ],
        }

        schema["properties"]["normalize"]["format"] = "checkbox"
        schema["properties"]["scale_vector_thickness"]["format"] = "checkbox"
        schema["properties"]["opacity"]["format"] = "range"
        schema["properties"]["opacity"]["step"] = 0.05
        return schema


class EnvironmentPreset(str, enum.Enum):
    none = "none"
    apartment = "apartment"
    city = "city"
    dawn = "dawn"
    forest = "forest"
    lobby = "lobby"
    night = "night"
    park = "park"
    studio = "studio"
    sunset = "sunset"
    warehouse = "warehouse"


class PathTracer(SettingsBase):
    """Experimental path tracer settings."""

    enabled: bool = False
    environment: EnvironmentPreset = EnvironmentPreset.studio
    metalness: float = Field(0.7, ge=0.0, le=1.0, description="Metalness")
    roughness: float = Field(0.2, ge=0.0, le=1.0, description="Roughness")
    clearcoat: float = Field(0.0, ge=0.0, le=1.0, description="Clearcoat")
    clearcoatRoughness: float = Field(
        0.0, ge=0.0, le=1.0, description="Clearcoat Roughness"
    )

    @classmethod
    def model_json_schema_from_atoms(cls, atoms: ase.Atoms) -> dict:
        schema = cls.model_json_schema()
        schema["properties"]["enabled"]["format"] = "checkbox"
        # make all of them sliders
        schema["properties"]["metalness"]["format"] = "range"
        schema["properties"]["roughness"]["format"] = "range"
        schema["properties"]["clearcoat"]["format"] = "range"
        schema["properties"]["clearcoatRoughness"]["format"] = "range"
        # also set the step
        schema["properties"]["metalness"]["step"] = 0.05
        schema["properties"]["roughness"]["step"] = 0.05
        schema["properties"]["clearcoat"]["step"] = 0.05
        schema["properties"]["clearcoatRoughness"]["step"] = 0.05
        return schema


SETTINGS = {
    Particle.__name__: Particle,
    Visualization.__name__: Visualization,
    Camera.__name__: Camera,
    PathTracer.__name__: PathTracer,
    VectorDisplay.__name__: VectorDisplay,
    Arrows.__name__: Arrows,
}



import logging
import typing as t

import socketio.exceptions
import znsocket.exceptions

from zndraw.base import Extension

if t.TYPE_CHECKING:
    from zndraw import ZnDraw


log = logging.getLogger(__name__)
TASK_RUNNING = "ZNDRAW TASK IS RUNNING"


def check_queue(vis: "ZnDraw") -> None:
    """Main loop to check and process modifier tasks for both private and public queues."""
    while True:
        if not vis._modifiers:
            vis.socket.sleep(1)
            continue
        try:
            process_modifier_queue(vis)
            process_public_queue(vis)
            vis.socket.sleep(1)
        except (znsocket.exceptions.ZnSocketError, socketio.exceptions.SocketIOError):
            log.warning("Connection to ZnDraw server lost. Reconnecting...")
            vis.socket.disconnect()
            vis.socket.sleep(1)


def process_modifier_queue(vis: "ZnDraw") -> None:
    """Process private modifier tasks in the queue."""
    modifier_queue = znsocket.Dict(
        r=vis.r,
        socket=vis._refresh_client,
        key=f"queue:{vis.token}:modifier",
    )

    for key in modifier_queue:
        if key in vis._modifiers:
            try:
                task = modifier_queue.pop(key)
                cls = vis._modifiers[key]["cls"]
                run_kwargs = vis._modifiers[key]["run_kwargs"]
                run_queued_task(vis, cls, task, modifier_queue, run_kwargs)
            except IndexError:
                pass


def process_public_queue(vis: "ZnDraw") -> None:
    """Process public modifier tasks in the public queue."""
    from zndraw import ZnDraw

    if not any(mod["public"] for mod in vis._modifiers.values()):
        return

    public_queue = znsocket.Dict(
        r=vis.r,
        socket=vis._refresh_client,
        key="queue:default:modifier",
    )

    for room, room_queue in public_queue.items():
        for key in room_queue:
            if key in vis._modifiers and vis._modifiers[key]["public"]:
                new_vis = ZnDraw(url=vis.url, token=room, r=vis.r)
                try:
                    task = room_queue.pop(key)
                    # run_queued_task(new_vis, key, task, room_queue)
                    cls = vis._modifiers[key]["cls"]
                    run_kwargs = vis._modifiers[key]["run_kwargs"]
                    run_queued_task(new_vis, cls, task, room_queue, run_kwargs)
                except IndexError:
                    pass
                finally:
                    new_vis.socket.sleep(1)
                    new_vis.socket.disconnect()


def run_queued_task(
    vis: "ZnDraw",
    cls: t.Type[Extension],
    task: dict,
    queue: znsocket.Dict,
    run_kwargs: dict | None = None,
) -> None:
    """Run a specific task and handle exceptions."""
    if not run_kwargs:
        run_kwargs = {}
    try:
        queue[TASK_RUNNING] = True
        cls(**task).run(vis, **run_kwargs)
    except Exception as err:
        vis.log(f"Error running `{cls}`: `{err}`")
    finally:
        queue.pop(TASK_RUNNING)



import functools
import logging
import pathlib
import socket
import typing as t
import urllib.parse
from urllib.parse import urlparse

import numpy as np
import plotly.graph_objects as go
import plotly.graph_objs
import socketio.exceptions
import znjson
from ase.data import covalent_radii

log = logging.getLogger(__name__)


def parse_url(input_url) -> t.Tuple[str, t.Optional[str]]:
    parsed = urlparse(input_url)
    base_url = f"{parsed.scheme}://{parsed.netloc}"
    path = parsed.path.strip("/") if parsed.path else None
    return base_url, path if path else None


def rgb2hex(value):
    r, g, b = np.array(value * 255, dtype=int)
    return "#%02x%02x%02x" % (r, g, b)


@functools.lru_cache(maxsize=128)
def get_scaled_radii() -> np.ndarray:
    """Scale down the covalent radii to visualize bonds better."""
    radii = covalent_radii
    # shift the values such that they are in [0.3, 1.3]
    radii = radii - np.min(radii)
    radii = radii / np.max(radii)
    radii = radii + 0.3
    return radii


def get_port(default: int) -> int:
    """Get an open port."""
    try:
        sock = socket.socket()
        sock.bind(("", default))
        port = default
    except OSError:
        sock = socket.socket()
        sock.bind(("", 0))
        port = sock.getsockname()[1]
    finally:
        sock.close()
    return port


class ZnDrawLoggingHandler(logging.Handler):
    """Logging handler which emits log messages to the ZnDraw server."""

    def __init__(self, zndraw):
        super().__init__()
        self.zndraw = zndraw

    def emit(self, record):
        try:
            msg = self.format(record)
            self.zndraw.log(msg)
        except RecursionError:  # See StreamHandler
            raise
        except Exception:
            print("Something went wrong")
            self.handleError(record)


def emit_with_retry(
    socket: socketio.Client,
    event,
    data=None,
    namespace=None,
    callback=None,
    retries: int = 1,
    delay: float = 0.1,
    increase_delay: float = 0.1,
    reconnect: bool = False,
) -> None:
    """Emit data to a socket with retries."""
    for idx in range(retries):
        try:
            socket.emit(event=event, data=data, namespace=namespace, callback=callback)
            break
        except socketio.exceptions.BadNamespaceError as err:
            log.error(f"Retrying {event} due to {err}")
            if idx == retries - 1:
                raise err
            if reconnect:
                raise ValueError("Reconnect not implemented yet")
        except Exception as err:
            if idx == retries - 1:
                raise err
        socket.sleep(delay)
        delay += increase_delay


def call_with_retry(
    socket: socketio.Client,
    event,
    data=None,
    namespace=None,
    timeout=60,
    retries: int = 1,
    delay: float = 0.1,
    increase_delay: float = 0.1,
    reconnect: bool = False,
) -> t.Any:
    """Call a function with retries."""
    for idx in range(retries):
        try:
            return socket.call(
                event=event, data=data, namespace=namespace, timeout=timeout
            )
        except (
            socketio.exceptions.TimeoutError,
            socketio.exceptions.BadNamespaceError,
        ) as err:
            log.error(f"Retrying {event} due to {err} ({idx} / {retries})")
            if idx == retries - 1:
                raise err
            if reconnect:
                raise ValueError("Reconnect not implemented yet")
        except Exception as err:
            if idx == retries - 1:
                raise err
        socket.sleep(delay)
        delay += increase_delay
    return None


def direction_to_euler(direction, roll=0):
    """
    Convert a direction vector to euler angles.

    You get an increased degree of freedom by setting the roll angle.
    """
    direction = np.array(direction)
    direction = direction / np.linalg.norm(direction)

    x, y, z = direction

    yaw = np.arctan2(y, x)
    pitch = np.arctan2(z, np.sqrt(x**2 + y**2))

    return np.array([yaw, pitch, roll])


def euler_to_direction(angles):
    """
    Convert euler angles to a direction vector.

    Roll gets discarded as this is a reduction in degrees of freedom.
    """
    yaw, pitch, roll = angles

    x = np.cos(yaw) * np.cos(pitch)
    y = np.sin(yaw) * np.cos(pitch)
    z = np.sin(pitch)

    return np.array([x, y, z])


def convert_url_to_http(url: str) -> str:
    """Convert a URL to a local file path."""
    url = urllib.parse.urlparse(url)
    if url.scheme == "wss":
        url = url._replace(scheme="https")
    elif url.scheme == "ws":
        url = url._replace(scheme="http")

    return urllib.parse.urlunparse(url)


def get_schema_with_instance_defaults(self) -> dict:
    """Update the schema defaults from the instance."""
    try:
        schema = self.get_updated_schema()
    except AttributeError:
        schema = self.model_json_schema()
    for key, value in self.__dict__.items():
        if key in schema["properties"]:
            schema["properties"][key]["default"] = value
    return schema


def get_plots_from_zntrack(path: str, remote: str | None, rev: str | None):
    node_name, attribute = path.split(".", 1)
    try:
        import os

        ## FIX for zntrack bug https://github.com/zincware/ZnTrack/issues/806
        import sys

        import zntrack

        sys.path.insert(0, os.getcwd())
        ##

        node = zntrack.from_rev(node_name, remote=remote, rev=rev)
        return getattr(node, attribute)
    except ImportError as err:
        raise ImportError(
            "You need to install ZnTrack to use the remote feature."
        ) from err


def load_plots_to_dict(
    paths: list[str], remote: str | None, rev: str | None
) -> dict[str, go.Figure]:
    data = {}
    for path in paths:
        if not pathlib.Path(path).exists():
            if remote is not None or rev is not None:
                plots = get_plots_from_zntrack(path, remote, rev)
            else:
                raise FileNotFoundError(f"File {path} does not")
        else:
            plots = znjson.loads(pathlib.Path(path).read_text())
        if isinstance(plots, plotly.graph_objs.Figure):
            data[path] = plots
        elif isinstance(plots, dict):
            if not all(isinstance(v, plotly.graph_objs.Figure) for v in plots.values()):
                raise ValueError("All values in the plots dict must be plotly.graph_objs")
            data.update({f"{path}_{k}": v for k, v in plots.items()})
        elif isinstance(plots, list):
            if not all(isinstance(v, plotly.graph_objs.Figure) for v in plots):
                raise ValueError("All values in the plots list must be plotly.graph_objs")
            data.update({f"{path}_{i}": v for i, v in enumerate(plots)})
        else:
            raise ValueError("The plots must be a dict, list or Figure")

    return data
